{-# LANGUAGE FunctionalDependencies #-}
module Types.FunctionIO
  ( IsFunctionIO(..)
  , ListDirectory(..)
  , IsDirectory(..)
  , ReadTextFile(..)
  , OpenImage(..)
  , SaveImage(..)
  , SaveTextFile(..)
  , CopyFile(..)
  , MakeDir(..)
  ) where

import Prelude hiding (String, FilePath)

import Types.Values
import Types.Value (Value(..), Valuable(..), makeImage)

import Data.Text (Text)
import qualified Data.Text.IO as T
import qualified Codec.Picture as CP
import System.Directory (listDirectory, doesDirectoryExist, createDirectory, copyFile)
import Language.Haskell.TH.Syntax (Lift)

class (Show f, Lift f, Valuable a, Valuable b) => IsFunctionIO f a b | f -> a b where
  evalFunctionIO :: f -> Value -> IO Value
  functionIOTouchesFilesystem :: f -> Bool

data ListDirectory = ListDirectory deriving (Show, Lift)
instance IsFunctionIO ListDirectory FilePath [FilePath] where
  evalFunctionIO ListDirectory = \case
    String (StringWrapper s) ->
      (List . map (toValue . StringWrapper)) <$> listDirectory s
    _ -> error "unexpected"
  functionIOTouchesFilesystem ListDirectory = False

data IsDirectory = IsDirectory deriving (Show, Lift)
instance IsFunctionIO IsDirectory FilePath Bool where
  evalFunctionIO IsDirectory = \case
    String (StringWrapper s) ->
      Bool <$> doesDirectoryExist s
    _ -> error "unexpected"
  functionIOTouchesFilesystem IsDirectory = False

data ReadTextFile = ReadTextFile deriving (Show, Lift)
instance IsFunctionIO ReadTextFile FilePath Text where
  evalFunctionIO ReadTextFile = \case
    String (StringWrapper s) ->
      Text <$> T.readFile s
    _ -> error "unexpected"
  functionIOTouchesFilesystem ReadTextFile = False

data OpenImage = OpenImage deriving (Show, Lift)
instance IsFunctionIO OpenImage FilePath Image where
  evalFunctionIO OpenImage = \case
    String (StringWrapper s) -> do
      imageOrig <- CP.readImage s
      case imageOrig of
        Left e -> error ("unexpected error: " ++ e)
        Right image -> pure $ makeImage $ CP.convertRGB8 image
    _ -> error "unexpected"
  functionIOTouchesFilesystem OpenImage = False

data SaveImage = SaveImage deriving (Show, Lift)
instance IsFunctionIO SaveImage (Image, FilePath) () where
  evalFunctionIO SaveImage = \case
    Tuple (Image (ImageWrapper image), String (StringWrapper s)) -> do
      CP.saveJpgImage 90 s $ CP.ImageRGB8 image
      pure Empty
    _ -> error "unexpected"
  functionIOTouchesFilesystem SaveImage = True

data SaveTextFile = SaveTextFile deriving (Show, Lift)
instance IsFunctionIO SaveTextFile (Text, FilePath) () where
  evalFunctionIO SaveTextFile = \case
    Tuple (Text t, String (StringWrapper s)) -> do
      T.writeFile s t
      pure Empty
    _ -> error "unexpected"
  functionIOTouchesFilesystem SaveTextFile = True

data CopyFile = CopyFile deriving (Show, Lift)
instance IsFunctionIO CopyFile (FilePath, FilePath) () where
  evalFunctionIO CopyFile = \case
    Tuple (String (StringWrapper source), String (StringWrapper target)) -> do
      copyFile source target
      pure Empty
    _ -> error "unexpected"
  functionIOTouchesFilesystem CopyFile = True

data MakeDir = MakeDir deriving (Show, Lift)
instance IsFunctionIO MakeDir FilePath () where
  evalFunctionIO MakeDir = \case
    String (StringWrapper s) -> do
      createDirectory s
      pure Empty
    _ -> error "unexpected"
  functionIOTouchesFilesystem MakeDir = True
